/**
 * This file is part of
 * 
 * Replicating Solidarity Evolution (ReSolEvo) 0.9
 *
 * Copyright (C) 2010 Center for Environmental Systems Research, Kassel, Germany
 * 
 * ReSolEvo is free software: you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * ReSolEvo is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Created by Sascha Holzhauer on 07.01.2010
 */
package de.cesr.resolevo;



import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.collections15.Transformer;
import org.apache.log4j.Logger;
import org.geotools.coverage.processing.operation.Resample;

import repast.simphony.context.Context;
import repast.simphony.dataLoader.ContextBuilder;
import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.engine.environment.RunState;
import repast.simphony.parameter.IllegalParameterException;
import repast.simphony.random.RandomHelper;
import repast.simphony.space.graph.RepastEdge;
import de.cesr.lara.components.LaraBehaviouralOption;
import de.cesr.lara.components.LaraEnvironment;
import de.cesr.lara.components.LaraGoal;
import de.cesr.lara.components.LaraPreprocessBuilder;
import de.cesr.lara.components.LaraPreprocessConfigurator;
import de.cesr.lara.components.impl.Log4jLogger;
import de.cesr.lara.components.impl.environment.LaraEnvironmentImpl;
import de.cesr.lara.components.impl.preprocessor.DefaultModeSelector;
import de.cesr.lara.components.impl.preprocessor.LPreprocessBuilder;
import de.cesr.lara.components.impl.preprocessor.LPseudoBoUpdaterBuilder;
import de.cesr.more.rs.adapter.DefaultLRsNetwork;
import de.cesr.resolevo.ReSolEvoModel.AmoreDecision;
import de.cesr.resolevo.ReSolEvoModel.MaximizePayoffGoal;
import de.cesr.resolevo.ReSolEvoParameterManager.Parameter;
import de.cesr.resolevo.agents.ReSolEvoAgent;
import de.cesr.resolevo.bos.ExploreAgent;
import de.cesr.resolevo.util.ReSolEvoParameterDB;
import edu.uci.ics.jung.io.GraphMLMetadata;
import de.cesr.resolevo.util.GraphMLWriter;



/**
 * AMoRe - Agents Managing ongoing Relations
 * 
 * @author Sascha Holzhauer
 * @date 22.02.2010
 * 
 */
public class ReSolEvoAgentContext implements ContextBuilder<ReSolEvoAgent> {

	/**
	 * Logger
	 */
	static private Logger loggerInitilNetwork = Log4jLogger.getLogger(ReSolEvoAgentContext.class.getName() + ".InitialNetwork");
	
	ReSolEvoParameterManager					paraMan;

	ReSolEvoNetwork		networkK;
	DefaultLRsNetwork<ReSolEvoAgent, RepastEdge<ReSolEvoAgent>>		lNetworkK;

	// required to initialise the network at end:
	static public Context<ReSolEvoAgent>	agentContext;

	@Override
	public Context<ReSolEvoAgent> build(Context<ReSolEvoAgent> context) {

		paraMan = ReSolEvoParameterDB.getInstance();
		networkK = new ReSolEvoNetwork(ReSolEvoModel.Network.NETWORK_KNOWN.toString(), context);

		context.addProjection(networkK);

		lNetworkK = new DefaultLRsNetwork<ReSolEvoAgent, RepastEdge<ReSolEvoAgent>>(networkK, context);

		RunState.getInstance().getMasterContext().findContext("AMoRePujol2005").add(networkK);

		createAgents(context);
		connectAgents(context);

		RunEnvironment.getInstance().getCurrentSchedule().schedule(this);
		agentContext = context;
		return context;
	}

	/**
	 * @param context
	 */
	public void createAgents(Context<ReSolEvoAgent> context) {
		LaraEnvironment env = new LaraEnvironmentImpl();

		// init preprocessor:
		LaraPreprocessConfigurator<ReSolEvoAgent, LaraBehaviouralOption<ReSolEvoAgent>> configurator = LPreprocessBuilder
				.getDefaultPreprocessConfigurator();

		configurator.setBOScanner(new ReSolEvoAgentSelectionBORecaller(), AmoreDecision.DBUILDER_SELECTION
				.getDBuilder());
		configurator.setModeSelector(new DefaultModeSelector<ReSolEvoAgent, LaraBehaviouralOption<ReSolEvoAgent>>(), AmoreDecision.DBUILDER_SELECTION
				.getDBuilder());
		configurator.setModeSelector(new ReSolEvoGameModeSelector<ReSolEvoAgent>());
		// Set PlayBoRecaller as default - so it is used for the numerous play decision builder:
		//configurator.setBOScanner(new ReSolEvoPlayBoRecaller());
		
		//configurator.setBOAdapter(new LPseudoBoUpdaterBuilder<ReSolEvoAgent, LaraBehaviouralOption>());

		LaraPreprocessBuilder<ReSolEvoAgent, LaraBehaviouralOption<ReSolEvoAgent>> ppBuilder = LPreprocessBuilder
				.<ReSolEvoAgent, LaraBehaviouralOption<ReSolEvoAgent>> getPreprocessBuilder(configurator);

		// agent initialisation:
		for (int i = 0; i < (Integer) paraMan.getParameter(Parameter.NUM_AGENTS); i++) {
			double attractiveness;
			switch (((Integer) ReSolEvoParameterDB.getInstance().getParameter(Parameter.ATTRACIVENESS_DIST))
					.intValue()) {
				case 1: {
					attractiveness = (double) Math
							.round(RandomHelper.getUniform().nextDoubleFromTo(0.05f, 0.95f) * 1000) / 1000;
					// Alternative: = RandomHelper.getUniform().nextIntFromTo(0, 1000) * 0.9 / 1000 + 0.05);
					break;
				}
				case 2: {
					attractiveness = RandomHelper.nextIntFromTo(0, 1) == 1 ? 
							(double) Math.round(RandomHelper.getNormal().nextDouble(1.0/3.0, 0.25) * 1000) / 1000 :
								(double) Math.round(RandomHelper.getNormal().nextDouble(2.0/3.0, 0.25) * 1000) / 1000;
					// Alternative: = RandomHelper.getUniform().nextIntFromTo(0, 1000) * 0.9 / 1000 + 0.05);
					break;
				}
				default:
					throw new IllegalParameterException("Parameter for distribution of attractivity is not supported");
			}
			ReSolEvoAgent agent = new ReSolEvoAgent(env, attractiveness,
			// +1 since the perceived own attractiveness is also placed in the memory (and every time step renewed not to be forgotten)
					(Integer) paraMan.getParameter(Parameter.INTERACTION_CAPACITY), (Integer) paraMan
							.getParameter(Parameter.MEMORY_CAPACITY) + 1);

			agent.setPreprocessBuilder(ppBuilder);

			Map<Class<? extends LaraGoal>, Double> preferences = new HashMap<Class<? extends LaraGoal>, Double>();
			preferences.put(MaximizePayoffGoal.class, new Double(1.0));
			agent.addPreferences(preferences);

			agent.getLaraNetworkComp().setNetwork(lNetworkK);
			context.add(agent);
		}
	}

	/**
	 * @param context
	 */
	/**
	 * @param context
	 *            Created by Sascha Holzhauer on 06.05.2010
	 */
	public void connectAgents(Context<ReSolEvoAgent> context) {

		ReSolEvoNetwork networkInitial = null;
		if (loggerInitilNetwork.isDebugEnabled()) {
			networkInitial = new ReSolEvoNetwork(
					ReSolEvoModel.Network.NETWORK_KNOWN.toString(), context);
		}
		
		List<ReSolEvoAgent> potNeighbours = new ArrayList<ReSolEvoAgent>();
		int random;

		// sort agents in list to get comparable results:
		List<ReSolEvoAgent> agents = new ArrayList<ReSolEvoAgent>();
		for (Object a : context.getAgentLayer(ReSolEvoAgent.class)) {
			agents.add((ReSolEvoAgent) a);
		}
		Collections.sort(agents, ReSolEvoAgent.getComparator());

		if (loggerInitilNetwork.isDebugEnabled()) {
			for (ReSolEvoAgent agent : agents) {
				networkInitial.addVertex(agent);
			}
		}
		for (ReSolEvoAgent agent : agents) {
			potNeighbours.clear();
			for (ReSolEvoAgent ag : context.getObjects(ReSolEvoAgent.class)) {
				potNeighbours.add(ag);
			}
			potNeighbours.remove(agent);
			// adaptation to original model:
			for (int i = 0; i < (Integer) paraMan.getParameter(Parameter.MEMORY_NUM_NOTICED_AGENTS) / 2; i++) {
				if (potNeighbours.size() == 0) {
					throw new IllegalStateException(
							"The current parameter combination of NUM_AGENTS and MEMORY_NUM_NOTICED_AGENTS "
									+ "and the current random seed do not allow a correct agent network initialisation");
				}
				// choose a random agent from the list of all other agents:
				ReSolEvoAgent alter = potNeighbours.get(random = RandomHelper.getUniform().nextIntFromTo(0,
						potNeighbours.size() - 1));
				potNeighbours.remove(random);

				// check whether the chosen agent is already a neighbour (because of a selection of that alter before):
				while (agent.getBOMemory().contains(alter.getAgentID() + ExploreAgent.EXPLORE_PREFIX)) {
					if (potNeighbours.isEmpty()) {
						throw new IllegalStateException(
								"The current parameter combination of NUM_AGENTS and MEMORY_NUM_NOTICED_AGENTS "
										+ "and the current random seed do not allow a correct agent network initialisation");
					}
					alter = potNeighbours.get(random = RandomHelper.getUniform().nextIntFromTo(0,
							potNeighbours.size() - 1));
					potNeighbours.remove(random);
				}
				// see Pujol2005, section 2.5
				agent.addNoticedAgent(alter, false);
				// adaptation to original model:
				alter.addNoticedAgent(agent, false);
				
				if (loggerInitilNetwork.isDebugEnabled()) {
					networkInitial.addEdge(agent, alter);
					networkInitial.addEdge(alter, agent);
				}
			}
		}
		if (loggerInitilNetwork.isDebugEnabled()) {
			loggerInitilNetwork.debug("Output initial graph...");
			File file1 = new File(((String)ReSolEvoParameterDB.getInstance().getParameter(Parameter.GRAPH_OUTPUT_DIR))
					 + ReSolEvoParameterDB.getInstance().getParameter(Parameter.PARAMS_ID) + System.getProperty("file.separator") + 
					"_" + "InitialNetwork" + ".graphml");
			if (!file1.getParentFile().exists()) {
				file1.getParentFile().mkdirs();
			}
			FileWriter fileWriter;
			try {
				fileWriter = new FileWriter(file1);
				GraphMLWriter<ReSolEvoAgent, RepastEdge<ReSolEvoAgent>> graphWriter = new GraphMLWriter<ReSolEvoAgent, RepastEdge<ReSolEvoAgent>>();
				
				// add attractiveness data:
				Map<String, GraphMLMetadata<ReSolEvoAgent>> attractivenessMap = new HashMap<String, GraphMLMetadata<ReSolEvoAgent>>();
				Transformer<ReSolEvoAgent, String> transformer = new Transformer<ReSolEvoAgent, String>() {
					public String transform(ReSolEvoAgent agent) {
						return new Double(agent.getAttractiveness()).toString();
					}
					
				};
				attractivenessMap.put("attractiveness", new GraphMLMetadata<ReSolEvoAgent>("attractiveness", "-1", transformer));
				graphWriter.setVertexData(attractivenessMap);
				
				graphWriter.save(networkInitial.getGraph(), fileWriter, "ReSolevoInitial");
			} catch (IOException e) {
				e.printStackTrace();
			}			
		}
	}

	/**
	 * @return the networkK
	 */
	public ReSolEvoNetwork getNetworkK() {
		return networkK;
	}
}
