/**
 * This file is part of
 * 
 * Replicating Solidarity Evolution (ReSolEvo) 0.9
 *
 * Copyright (C) 2010 Center for Environmental Systems Research, Kassel, Germany
 * 
 * ReSolEvo is free software: you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * ReSolEvo is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Created by Sascha Holzhauer on 07.01.2010
 */
package de.cesr.resolevo;



import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;

import org.apache.commons.collections15.Transformer;
import org.apache.log4j.Logger;

import repast.simphony.context.Context;
import repast.simphony.parameter.IllegalParameterException;
import repast.simphony.space.graph.DirectedJungNetwork;
import repast.simphony.space.graph.RepastEdge;
import repast.simphony.space.graph.UndirectedJungNetwork;
import de.cesr.lara.components.LaraBehaviouralOption;
import de.cesr.lara.components.impl.LModel;
import de.cesr.lara.components.impl.Log4jLogger;
import de.cesr.resolevo.ReSolEvoParameterManager.Parameter;
import de.cesr.resolevo.agents.ReSolEvoAgent;
import de.cesr.resolevo.bos.AskKnownAgent;
import de.cesr.resolevo.bos.SelectAgent;
import de.cesr.resolevo.util.GraphMLWriter;
import de.cesr.resolevo.util.ReSolEvoParameterDB;
import edu.uci.ics.jung.graph.DirectedSparseGraph;
import edu.uci.ics.jung.graph.Graph;
import edu.uci.ics.jung.graph.Hypergraph;
import edu.uci.ics.jung.graph.UndirectedGraph;
import edu.uci.ics.jung.graph.UndirectedSparseGraph;
import edu.uci.ics.jung.graph.util.EdgeType;
import edu.uci.ics.jung.graph.util.Pair;
import edu.uci.ics.jung.io.GraphMLMetadata;
import edu.uos.sh.soneta.measures.supply.network.AverageShortestPath;
import edu.uos.sh.soneta.measures.supply.network.ClusteringCoefficient;
import edu.uos.sh.soneta.sadapted.context.space.graph.ContextJungNetworkAdapted;



/**
 * AMoRe - Agents Managing ongoing Relations
 * 
 * @author Sascha Holzhauer
 * @param <AgentType> Type of agents within the network
 * @date 07.01.2010
 */
public class ReSolEvoNetwork extends ContextJungNetworkAdapted<ReSolEvoAgent> {

	static private Logger	logger				= Log4jLogger.getLogger(ReSolEvoNetwork.class);

	static int				directedGraphAge	= Integer.MIN_VALUE;
	static boolean			initialized			= false;

	static long				startTimeMillis		= 0;

	/**
	 * @param name the network's name
	 * @param context the context this network resides in
	 */
	public ReSolEvoNetwork(String name, Context<ReSolEvoAgent> context) {
		super(new DirectedJungNetwork<ReSolEvoAgent>(name), context);
	}

	public void initNetwork() {
		startTimeMillis = System.currentTimeMillis();
		int step = LModel.getModel().getCurrentStep();
		if (directedGraphAge < step) {
			if (logger.isDebugEnabled()) {
				logger.debug(this.getName() + "> init Network");
			}

			setGraph(new DirectedSparseGraph<ReSolEvoAgent, RepastEdge<ReSolEvoAgent>>());
			for (ReSolEvoAgent agent : ReSolEvoAgentContext.agentContext) {
				addVertex(agent);
			}
			for (ReSolEvoAgent agent : ReSolEvoAgentContext.agentContext) {
				for (LaraBehaviouralOption bo : agent.getBOMemory().recallAllMostRecent()) {
					ReSolEvoAgent partner = ((SelectAgent) bo).getAlter();
					// only add an edge when the cooperation is benefical for both (i.e. positive outcome):
					// > both agents cooperate with positive outcome if and only if the net benefit is positive for each
					// of the partners
					if (bo instanceof AskKnownAgent) {
						if (!isAdjacent(agent, partner)) {
							addEdge(agent, partner);
						}
					}
				}
			}
			directedGraphAge = step; // LModel.getModel().getCurrentStep();
			outputGraph();
			if (logger.isDebugEnabled()) {
				try {
					logger.debug(this.getName() + "> initialisation finished (taking "
							+ DatatypeFactory.newInstance().newDuration((System.currentTimeMillis() - startTimeMillis))
							+ ")");
				} catch (DatatypeConfigurationException e) {
					e.printStackTrace();
				}
			}
		}
	}

	public Graph getOptimalNetwork() {
		startTimeMillis = System.currentTimeMillis();
		UndirectedGraph<ReSolEvoAgent, Object> optimalGraph = new UndirectedSparseGraph<ReSolEvoAgent, Object>();

		int cont, first = 0;

		int numAgents = ((Integer) ReSolEvoParameterDB.getInstance().getParameter(Parameter.NUM_AGENTS)).intValue();
		int numCapacity = ((Integer) ReSolEvoParameterDB.getInstance().getParameter(Parameter.INTERACTION_CAPACITY))
				.intValue();
		ReSolEvoAgent[] agents = new ReSolEvoAgent[numAgents];
		int i = 0;
		for (ReSolEvoAgent agent : ReSolEvoAgentContext.agentContext) {
			agents[i++] = agent;
			optimalGraph.addVertex(agent);
		}

		for (i = 0; i < numAgents; i++) {
			cont = first;
			for (int k = 0; k < ((Integer) ReSolEvoParameterDB.getInstance().getParameter(
					Parameter.NUM_INTERACTIONS_AT_ONCE)).intValue(); k++) {
				for (int j = cont; j < numAgents; j++) {
					if (i != j) {
						// SH How could the degree exceed capacity ? - only when mutual relations are considered
						if ((optimalGraph.degree(agents[j]) < numCapacity)
								&& optimalGraph.isNeighbor(agents[i], agents[j])) { // SH edge could have been created
																					// before...
							if (agents[i].getReceivingAdviceGain(agents[j]) > agents[i].getGivingAdviseLoss(agents[j])
									&& agents[j].getReceivingAdviceGain(agents[i]) > agents[j]
											.getGivingAdviseLoss(agents[i])) {
								optimalGraph.addEdge(new Object(), agents[i], agents[j]);
							}
							cont = j + 1;
							if ((k == 0) && (i > 100))
								first = j; // SH ??
							break;
						}
					}
				}
			}
		}
		if (logger.isDebugEnabled()) {
			try {
				logger.debug(this.getName() + "> caclulate optimal network (taking "
						+ DatatypeFactory.newInstance().newDuration((System.currentTimeMillis() - startTimeMillis))
						+ ")");
			} catch (DatatypeConfigurationException e) {
				e.printStackTrace();
			}
		}
		return optimalGraph;
	}

	/**
	 * Returns the average shortest path between all pairs of nodes
	 * 
	 * @return Created by Sascha Holzhauer on 14.04.2010
	 */
	public double getAvgPathLength() {
		initNetwork();
		startTimeMillis = System.currentTimeMillis();
		// double value = AverageShortestPath.getAverageShortestPathR(getUndirected());
		double value = AverageShortestPath.getAverageShortestPathUnweighted(getUndirected());
		if (logger.isDebugEnabled()) {
			try {
				logger.debug(this.getName()
						+ "> calculation of average path length - getAverageShortestPathUnweighted (taking "
						+ DatatypeFactory.newInstance().newDuration((System.currentTimeMillis() - startTimeMillis))
						+ ")");
			} catch (DatatypeConfigurationException e) {
				e.printStackTrace();
			}
		}
		return value;
	}

	/**
	 * Returns the clustering coefficient as implemented in R
	 * 
	 * @return Created by Sascha Holzhauer on 14.04.2010
	 */
	public double getClusteringCoefficient() {
		initNetwork();
		startTimeMillis = System.currentTimeMillis();
		if (logger.isDebugEnabled()) {
			logger.debug(this.getName() + "> calculate CC # 1");
		}
		double value = Double.NaN;
		switch (((Integer) ReSolEvoParameterDB.getInstance().getParameter(Parameter.CC_ALGO)).intValue()) {
			case 1:
			case 4:
			case 6:
			case 7:
				value = ClusteringCoefficient.getClusteringCoefficientR(getUndirected());
				break;
			case 2:
			case 3:
			case 5:
				value = 0.0;
				break;
			default:
				throw new IllegalParameterException(
						"Parameter for calculating the clustering coefficient not supported");
		}
		if (logger.isDebugEnabled()) {
			try {
				logger.debug(this.getName() + "> calculate CC # 1 (taking "
						+ DatatypeFactory.newInstance().newDuration((System.currentTimeMillis() - startTimeMillis))
						+ ")");
			} catch (DatatypeConfigurationException e) {
				e.printStackTrace();
			}
		}
		return value;

	}

	public double getClusteringCoefficient2() {
		initNetwork();
		startTimeMillis = System.currentTimeMillis();
		if (logger.isDebugEnabled()) {
			logger.debug(this.getName() + "> calculate CC # 2");
		}
		double value = Double.NaN;
		switch (((Integer) ReSolEvoParameterDB.getInstance().getParameter(Parameter.CC_ALGO)).intValue()) {
			case 2:
			case 4:
			case 5:
			case 7:
				value = ClusteringCoefficient.getClusteringCoefficientPerNode(getUndirected());
				break;
			case 1:
			case 3:
			case 6:
				value = 0.0;
				break;
			default:
				throw new IllegalParameterException(
						"Parameter for calculating the clustering coefficient not supported");
		}
		if (logger.isDebugEnabled()) {
			try {
				logger.debug(this.getName() + "> calculate CC # 2 (taking "
						+ DatatypeFactory.newInstance().newDuration((System.currentTimeMillis() - startTimeMillis))
						+ ")");
			} catch (DatatypeConfigurationException e) {
				e.printStackTrace();
			}
		}
		return value;
	}

	public double getClusteringCoefficient3() {
		initNetwork();
		startTimeMillis = System.currentTimeMillis();
		if (logger.isDebugEnabled()) {
			logger.debug(this.getName() + "> calculate CC # 3");
		}
		double value = Double.NaN;
		switch (((Integer) ReSolEvoParameterDB.getInstance().getParameter(Parameter.CC_ALGO)).intValue()) {
			case 3:
			case 5:
			case 6:
			case 7:
				value = ClusteringCoefficient.getClusteringCoefficientPerNodeWithNeighbors(getUndirected());
				break;
			case 1:
			case 2:
			case 4:
				value = 0.0;
				break;
			default:
				throw new IllegalParameterException(
						"Parameter for calculating the clustering coefficient not supported");
		}
		if (logger.isDebugEnabled()) {
			try {
				logger.debug(this.getName() + "> calculate CC # 3 (taking "
						+ DatatypeFactory.newInstance().newDuration((System.currentTimeMillis() - startTimeMillis))
						+ ")");
			} catch (DatatypeConfigurationException e) {
				e.printStackTrace();
			}
		}
		return value;
	}

	/**
	 * Returns the current parameter id (required for outputter to be defined in this class)
	 * 
	 * @return Created by Sascha Holzhauer on 14.04.2010
	 */
	public int getParamID() {
		return ((Integer) ReSolEvoParameterDB.getInstance().getParameter(Parameter.PARAMS_ID)).intValue();
	}

	/**
	 * TODO test
	 * 
	 * @return Created by Sascha Holzhauer on 30.04.2010
	 */
	@SuppressWarnings("unchecked")
	public Graph getUndirected() {
		initNetwork();

		if (this.getGraph() instanceof UndirectedGraph) {
			return this.getGraph();
		} else {
			startTimeMillis = System.currentTimeMillis();
			Graph<ReSolEvoAgent, RepastEdge<ReSolEvoAgent>> graph = this.getGraph();
			UndirectedGraph out = new UndirectedSparseGraph<ReSolEvoAgent, RepastEdge<ReSolEvoAgent>>();

			for (ReSolEvoAgent v : graph.getVertices())
				out.addVertex(v);

			for (RepastEdge<ReSolEvoAgent> e : graph.getEdges()) {
				Pair<ReSolEvoAgent> endpoints = graph.getEndpoints(e);
				// ??
				if (graph.isSuccessor(endpoints.getFirst(), endpoints.getSecond())
						|| graph.isPredecessor(endpoints.getFirst(), endpoints.getSecond())) {
					if (!out.isNeighbor(endpoints.getFirst(), endpoints.getSecond())) {
						out.addEdge(e, endpoints.getFirst(), endpoints.getSecond(), EdgeType.UNDIRECTED);
					}
				}
			}
			this.setGraph(out);
			if (logger.isDebugEnabled()) {
				try {
					logger.debug(this.getName() + "> get undirected (taking "
							+ DatatypeFactory.newInstance().newDuration((System.currentTimeMillis() - startTimeMillis))
							+ ")");
				} catch (DatatypeConfigurationException e) {
					e.printStackTrace();
				}
			}
		}
		return getGraph();
	}

	/**
	 * TODO handle exception Created by Sascha Holzhauer on 06.05.2010
	 */
	public void outputGraph() {
		startTimeMillis = System.currentTimeMillis();

		int timeStep = LModel.getModel().getCurrentStep();

		if (logger.isDebugEnabled()) {
			logger.debug(this.getName() + "> generate output");
		}
		File file1 = new File(((String) ReSolEvoParameterDB.getInstance().getParameter(Parameter.GRAPH_OUTPUT_DIR))
				+ ReSolEvoParameterDB.getInstance().getParameter(Parameter.PARAMS_ID)
				+ System.getProperty("file.separator") + LModel.getModel().getIntegerFormat().format(timeStep)
				+ System.getProperty("file.separator") + "_" + ReSolEvoModel.Network.NETWORK_KNOWN.toString()
				+ ".graphml");

		File file2 = new File(((String) ReSolEvoParameterDB.getInstance().getParameter(Parameter.GRAPH_OUTPUT_DIR))
				+ ReSolEvoParameterDB.getInstance().getParameter(Parameter.PARAMS_ID)
				+ System.getProperty("file.separator") + LModel.getModel().getIntegerFormat().format(timeStep)
				+ System.getProperty("file.separator") + "_" + ReSolEvoModel.Network.NETWORK_KNOWN.toString()
				+ "OptimalNet" + ".graphml");
		try {
			if (!file1.getParentFile().exists()) {
				file1.getParentFile().mkdirs();
			}
			if (!file2.getParentFile().exists()) {
				file2.getParentFile().mkdirs();
			}
			FileWriter fileWriter = new FileWriter(file1);
			GraphMLWriter<ReSolEvoAgent, RepastEdge<ReSolEvoAgent>> graphWriter = new GraphMLWriter<ReSolEvoAgent, RepastEdge<ReSolEvoAgent>>();

			if (logger.isDebugEnabled()) {
				logger.debug(this.getName() + "> save " + ReSolEvoModel.Network.NETWORK_KNOWN.toString()
						+ " (interaction)");
			}

			// add attractiveness data:
			Map<String, GraphMLMetadata<ReSolEvoAgent>> attractivenessMap = new HashMap<String, GraphMLMetadata<ReSolEvoAgent>>();
			Transformer<ReSolEvoAgent, String> transformer = new Transformer<ReSolEvoAgent, String>() {
				public String transform(ReSolEvoAgent agent) {
					return new Double((agent).getAttractiveness()).toString();
				}

			};
			attractivenessMap.put("attractiveness", new GraphMLMetadata<ReSolEvoAgent>("attractiveness", "-1",
					transformer));
			graphWriter.setVertexData(attractivenessMap);

			// add graph id:
			Map<String, GraphMLMetadata<Hypergraph<ReSolEvoAgent, RepastEdge<ReSolEvoAgent>>>> graphMap = new HashMap<String, GraphMLMetadata<Hypergraph<ReSolEvoAgent, RepastEdge<ReSolEvoAgent>>>>();
			Transformer<Hypergraph<ReSolEvoAgent, RepastEdge<ReSolEvoAgent>>, String> gTransformer = new Transformer<Hypergraph<ReSolEvoAgent, RepastEdge<ReSolEvoAgent>>, String>() {
				public String transform(Hypergraph<ReSolEvoAgent, RepastEdge<ReSolEvoAgent>> graph) {
					return "ResolevoInteractionGraph";
				}

			};
			graphMap.put("id", new GraphMLMetadata<Hypergraph<ReSolEvoAgent, RepastEdge<ReSolEvoAgent>>>("id", "-1",
					gTransformer));
			graphWriter.setGraphData(graphMap);

			graphWriter.save(this.getGraph(), fileWriter, "ResolevoInteraction");

			FileWriter fileWriter2 = new FileWriter(file2);
			GraphMLWriter<ReSolEvoAgent, RepastEdge<ReSolEvoAgent>> graphWriter2 = new GraphMLWriter<ReSolEvoAgent, RepastEdge<ReSolEvoAgent>>();

			if (logger.isDebugEnabled()) {
				logger.debug(this.getName() + "> save " + ReSolEvoModel.Network.NETWORK_KNOWN.toString()
						+ " (optimalNet)");
			}
			graphWriter2.save(getOptimalNetwork(), fileWriter2, "ReSolevoOptimal");
		} catch (IOException e) {
			e.printStackTrace();
			logger.error(this.getName() + "> generation of output failed (" + file1 + "/" + file2 + ")");
		}
		if (logger.isDebugEnabled()) {
			try {
				logger.debug(this.getName() + "> output graph (taking "
						+ DatatypeFactory.newInstance().newDuration((System.currentTimeMillis() - startTimeMillis))
						+ ")");
			} catch (DatatypeConfigurationException e) {
				e.printStackTrace();
			}
		}
	}

	/**
	 * Resets the graph age which determines if the graph needs to be re-initiated. Created by Sascha Holzhauer on
	 * 18.08.2010
	 */
	public static void reset() {
		directedGraphAge = Integer.MIN_VALUE;
	}
}
