/**
 * This file is part of
 * 
 * Replicating Solidarity Evolution (ReSolEvo) 0.9
 *
 * Copyright (C) 2010 Center for Environmental Systems Research, Kassel, Germany
 * 
 * ReSolEvo is free software: you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * ReSolEvo is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Created by Sascha Holzhauer on 07.01.2010
 */
package de.cesr.resolevo;


/**
 * 
 * Defines an interface for classes that provide parameter values. Also defines
 * all parameters used throughout the model.
 * 
 * @author Holzhauer
 * @date 08.01.2009
 * 
 */
public interface ReSolEvoParameterManager {

	/**
	 * @author Sascha Holzhauer
	 * @date 29.03.2010
	 * 
	 */
	enum Parameter {

		/**
		 * Database Settings:
		 */
		SQL_LOCATION(String.class, "mysql2"),
		SQL_DBNAME(String.class, "klimzug"),
		SQL_USER(String.class, "holzhauer"),
		SQL_PASSWORD(String.class, "ymxncbv"),
		SQL_TBLNAME_PARAMS(String.class, "amorepujol2005_params"),

		PARAMS_ID(Integer.class, 21),
		
		/**
		 * Directory where the graphml-file is stored.
		 */
		GRAPH_OUTPUT_DIR(String.class, "./graphs"),
		
		/**
		 * random seed for RS random number generator
		 */
		PARAMS_RANDOMSEED(Integer.class, 10),
		
		/**
		 * Number of agent entries an agent is able to handle at a certain time
		 * step. Corresponds to Pujol's M_c
		 */
		MEMORY_CAPACITY(Integer.class, 5),

		/**
		 * Number of agents that are initially represented as unknowns in memory
		 * Corresponds to Pujol's M_0
		 */
		MEMORY_NUM_NOTICED_AGENTS(Integer.class, 4),

		/**
		 * overall number of agents
		 */
		NUM_AGENTS(Integer.class, 10),

		/**
		 * number of interactions an agent may establish during a time step
		 * 
		 */
		NUM_INTERACTIONS_AT_ONCE(Integer.class, 1),
		
		/**
		 * number of interactions an agent may establish during a time step
		 * Corresponds to Pujol's Q
		 */
		NUM_INTERACTIONS(Integer.class, 3),
		
		/**
		 * The number of interactions an agent my handle
		 */
		INTERACTION_CAPACITY(Integer.class, 150),
		
		/**
		 * Method to count current interactions
		 * 1 = as in pseudo code: init Q interactions no matter of current interactions but refuse interaction if current interactions exceed C
		 * 2 = as in description: count all interactions
		 */
		INTERACTION_CAPACITY_METHOD(Integer.class, 1),

		/**
		 * corresponds to B (B/E >! 1)
		 */
		ADVICE_BENEFIT(Double.class, 16.0),

		/**
		 * corresponds to E (B/E >! 1)
		 */
		ADVICE_COST(Double.class, 3.0),
		
		/**
		 * Algorithm to use for calculating
		 * the clustering coefficient
		 * A = R implementation
		 * B = per node
		 * C = per node with neighbours
		 * 1 = A, 2 = B, 3 = C
		 * 4 = A & B
		 * 5 = B & C
		 * 6 = A & C
		 * 7 = A & B & C
		 */
		CC_ALGO(Integer.class, 1),
		
		/**
		 * Method of assigning attractiveness:
		 * 1 = as in Pujol2005
		 */
		ATTRACIVENESS_DIST(Integer.class, 1),
		
		/**
		 * The mechanism of memory exchange.
		 * 1 = explicit,
		 * 2 = implicit memory exchange.
		 * 3 = both explicit and implicit memory exchange is applied.
		 */
		MEMORY_EXCHANGE(Integer.class, 1),
		
		/**
		 * Mode of consequences through memory exchange
		 * 1 = exchanged agents resident in know's network
		 * 2 = exchanged agents resident in noticed's network
		 * 3 = gossip > know's network, observation > noticed's network
		 */
		MEMORY_EXCHANGE_CONSEQUENCE(Integer.class, 1),
		
		/**
		 * Among which agents does an agent explore?
		 * 1 = all known and noticed  (description and pseudo code)
		 * 2 = only noticed
		 */
		EXPLORATION_METHOD(Integer.class, 1),
		
		/**
		 * Calculation method of exploration probability:
		 * 1 = consider not known partners from ME
		 * 2 = consider all known partners
		 */
		EXPLORATION_PROPABILITY(Integer.class, 1),
		
		/**
		 * 1 = version of pseudo code: accept when j is not aware of i or when expected payoff is non-negative in case j is aware of i 
		 * 2 = description version: j will only accept the interaction if [...] j is aware of i's existence (does not work!)
		 */
		INTERACTION_ACCEPTANCE_METHOD(Integer.class, 1),
		
		/**
		 * Order of agent execution (mainly regarding updating payoffs)
		 * 1 = iterative (but two decision stages [agent selection and game play])
		 * 2 = parallel/synchronised
		 * 3 = as in pseudo code: iterative (entire decision process including agent selection and game play)
		 */
		AGENT_TRIGGERING(Integer.class, 1),
		
		/**
		 * the simulation duration
		 */
		END_TICK(Integer.class, 100);
		

		private Class<?> type;
		private Object defaultValue;

		Parameter(Class<?> type) {
			this(type, null);
		}

		Parameter(Class<?> type, Object defaultValue) {
			this.type = type;
			this.defaultValue = defaultValue;
		}

		public Class<?> getType() {
			return type;
		}

		public Object getDefaultValue() {
			return defaultValue;
		}
	}

	/**
	 * Initializes the parameters like reading them from a file or instantiating
	 * RS's Parameters class.
	 * 
	 * Created by Holzhauer on 08.01.2009
	 */
	public void initParameters();

	/**
	 * Get any registered parameter
	 * 
	 * @param parameter
	 *            the {@link Parameter} whose value is requested
	 * @return the parameter's current value
	 * 
	 *         Created by Holzhauer on 08.01.2009
	 */
	public Object getParameter(Parameter parameter);
}
