/**
 * This file is part of
 * 
 * Replicating Solidarity Evolution (ReSolEvo) 0.9
 *
 * Copyright (C) 2010 Center for Environmental Systems Research, Kassel, Germany
 * 
 * ReSolEvo is free software: you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * ReSolEvo is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Created by Sascha Holzhauer on 07.01.2010
 */
package de.cesr.resolevo.agents;



import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.collections15.BidiMap;
import org.apache.commons.collections15.bidimap.DualTreeBidiMap;
import org.apache.log4j.Logger;

import repast.simphony.context.space.graph.ContextJungNetwork;
import repast.simphony.parameter.IllegalParameterException;
import repast.simphony.random.RandomHelper;
import repast.simphony.space.graph.RepastEdge;
import de.cesr.lara.components.LaraAgentComponent;
import de.cesr.lara.components.LaraBehaviouralOption;
import de.cesr.lara.components.LaraCapacityManageable;
import de.cesr.lara.components.LaraCapacityManagementView;
import de.cesr.lara.components.LaraCapacityManager;
import de.cesr.lara.components.LaraDecisionBuilder;
import de.cesr.lara.components.LaraDecisionData;
import de.cesr.lara.components.LaraEnvironment;
import de.cesr.lara.components.LaraGoal;
import de.cesr.lara.components.LaraMemoryPropertyListener;
import de.cesr.lara.components.LaraProperty;
import de.cesr.lara.components.LaraSimpleAgent;
import de.cesr.lara.components.impl.LModel;
import de.cesr.lara.components.impl.Log4jLogger;
import de.cesr.lara.components.impl.container.memory.DefaultLLimitedCapacityOverwriteBoMemory;
import de.cesr.lara.components.impl.decision.LDecisionHeuristicComponent_MaxLineTotalRandomAtTie;
import de.cesr.more.lara.AbstractLaraNetworkAgent;
import de.cesr.more.rs.adapter.DefaultLRsNetwork;
import de.cesr.resolevo.ReSolEvoModel;
import de.cesr.resolevo.ReSolEvoNetwork;
import de.cesr.resolevo.ReSolEvoParameterManager;
import de.cesr.resolevo.ReSolEvoModel.AmoreDecision;
import de.cesr.resolevo.ReSolEvoModel.MaximizePayoffGoal;
import de.cesr.resolevo.ReSolEvoParameterManager.Parameter;
import de.cesr.resolevo.bos.AskKnownAgent;
import de.cesr.resolevo.bos.AvoidSupportGame;
import de.cesr.resolevo.bos.ExploreAgent;
import de.cesr.resolevo.bos.PlaySupportGame;
import de.cesr.resolevo.bos.SelectAgent;
import de.cesr.resolevo.util.AgentLevel;
import de.cesr.resolevo.util.ReSolEvoParameterDB;
import edu.uos.sh.soneta.core.change.ChangeFiring;
import edu.uos.sh.soneta.measures.NetworkMeasureSupport;
import edu.uos.sh.soneta.measures.NetworkMeasureUtilities.MeasureDescription;



/**
 * AMoRe - Agents Managing ongoing Relations
 * 
 * @author Sascha Holzhauer
 * @date 05.01.2010
 * 
 */
public class ReSolEvoAgent extends
		AbstractLaraNetworkAgent<ReSolEvoAgent, RepastEdge<ReSolEvoAgent>, LaraBehaviouralOption<ReSolEvoAgent>>
		implements LaraMemoryPropertyListener, Comparable<ReSolEvoAgent>, ChangeFiring, NetworkMeasureSupport {

	/**
	 * Logger
	 */
	static private Logger		logger	= Log4jLogger.getLogger(ReSolEvoAgent.class);
	private Logger				agentlogger;

	private double				attractiveness;

	ReSolEvoParameterManager	paraMan;

	/**
	 * @return the attractiveness
	 */
	public double getAttractiveness() {
		return this.attractiveness;
	}

	private final int											interaction_capacity;

	private int													numCurrentInteractions	= 0;

	// TODO make more efficient (?):
	private BidiMap<ReSoLEvoPlayDecisionBuilder, ReSolEvoAgent>	gameDecisionsActive;
	private BidiMap<ReSoLEvoPlayDecisionBuilder, ReSolEvoAgent>	gameDecisionsPassive;

	private int													numNoticed				= 0;
	private int													numKnown				= 0;

	/**
	 * Partners added as knowns through memory exchange
	 */
	private int													numKnownME				= 0;

	Map<String, SelectAgent>									usedBOs					= new HashMap<String, SelectAgent>();

	public void addUsedBO(SelectAgent selectAgent) {
		usedBOs.put(selectAgent.getKey(), selectAgent);
	}

	public Map<String, SelectAgent> getUsedBOs() {
		return usedBOs;
	}

	/**
	 * BO-Memory capacity manager that removes the BO with lowest absolute(!) expected payoff (goal
	 * ReSolEvoModel.MaximizePayoffGoal). In case there are more than one entry, one is chosen at random.
	 * 
	 * AMoRe_Pujol2005
	 * 
	 * @author Sascha Holzhauer
	 * @date 29.04.2010
	 * 
	 */
	private static class PayoffCapacityManager implements LaraCapacityManager<LaraBehaviouralOption<ReSolEvoAgent>> {

		@Override
		public boolean manage(LaraCapacityManageable<LaraBehaviouralOption<ReSolEvoAgent>> capacityManageable) {

			LaraCapacityManagementView<LaraBehaviouralOption<ReSolEvoAgent>> view = capacityManageable
					.getCapacityManagementView();

			if (view == null)
				return false;

			double propValue;

			List<LaraBehaviouralOption<ReSolEvoAgent>> leastAttractiveBOs = new ArrayList<LaraBehaviouralOption<ReSolEvoAgent>>();
			LaraBehaviouralOption<ReSolEvoAgent> leastAttractiveProperty = null;
			double leastAttractivePropertyValue = Double.POSITIVE_INFINITY;

			for (LaraBehaviouralOption<ReSolEvoAgent> prop : view) {
				if (prop instanceof SelectAgent) {
					SelectAgent property = (SelectAgent) prop;
					if (property instanceof ExploreAgent) {
						propValue = 0.0;
					} else {
						propValue = Math.abs(property.getValue().get(ReSolEvoModel.MaximizePayoffGoal.class));
					}
					if (leastAttractiveProperty == null || propValue < leastAttractivePropertyValue) {
						leastAttractiveProperty = property;
						leastAttractivePropertyValue = propValue;
						leastAttractiveBOs.clear();
						leastAttractiveBOs.add(leastAttractiveProperty);
					} else {
						if (propValue < leastAttractivePropertyValue)
							leastAttractiveBOs.add(property);
					}
				}
			}
			if (leastAttractiveProperty != null) {
				view.remove(leastAttractiveBOs.get(RandomHelper.nextIntFromTo(0, leastAttractiveBOs.size() - 1)));
				return true;
			}
			return false;
		}
	}

	/**
	 * @param env
	 * @param attractiveness
	 * @param interaction_capacity the number of interaction this agent may encounter during one time step
	 * @param memoryCapacity
	 * 
	 */
	public ReSolEvoAgent(LaraEnvironment env, double attractiveness, int interaction_capacity, int memoryCapacity) {
		super(env);

		// init agent specific logger:
		if (Log4jLogger.getLogger(ReSolEvoAgent.class.getName() + "." + getAgentID()).isEnabledFor(AgentLevel.AGENT)) {
			agentlogger = Log4jLogger.getLogger(ReSolEvoAgent.class.getName() + "." + getAgentID());
		}

		this.interaction_capacity = interaction_capacity;
		gameDecisionsActive = new DualTreeBidiMap<ReSoLEvoPlayDecisionBuilder, ReSolEvoAgent>();
		gameDecisionsPassive = new DualTreeBidiMap<ReSoLEvoPlayDecisionBuilder, ReSolEvoAgent>();

		paraMan = ReSolEvoParameterDB.getInstance();

		this.attractiveness = attractiveness;

		// set up BO memory:
		setBOMemory(new DefaultLLimitedCapacityOverwriteBoMemory<LaraBehaviouralOption<ReSolEvoAgent>>(
				new PayoffCapacityManager(), memoryCapacity, "BOMemory.(" + this.getAgentID() + ")"));

		// setBOMemory(new DefaultLLimitedCapacityBOMemory<LaraBehaviouralOption>(new PayoffCapacityManager(),
		// memoryCapacity, "BOMemory.(" + this.getAgentID() + ")"));

		// to synchronise BO - memory and network:
		getBOMemory().addMemoryPropertyObserver(LaraMemoryPropertyListener.MemoryPropertyEvent.ITEM_FORGOTTEN, this);
		// getBOMemory().addMemoryPropertyObserver(LaraMemoryPropertyListener.MemoryPropertyEvent.REFRESHED_ITEM_FORGOTTEN,
		// this);

		// UNDO
		// getBOMemory().addMemoryPropertyObserver(LaraMemoryPropertyListener.MemoryPropertyEvent.ITEM_MEMORIZED, this);

		logger.info("Agent " + getAgentID() + " instantiated.");
		if (agentlogger != null) {
			agentlogger.log(AgentLevel.AGENT, getAgentID() + "> instantiated with attractiveness " + attractiveness);
		}
	}

	/**
	 * This method is executed right after the decision is made.
	 * 
	 * @see de.cesr.lara.components.impl.AbstractLaraAgent#afterDecide(LaraDecisionBuilder)
	 */
	@Override
	public void afterDecide(LaraDecisionBuilder dBuilder) {
		logger.info(getAgentID() + "> afterDecide");
		// <- LOGGING
		// if (logger.isDebugEnabled()) {
		// logger.debug("Number of Kown: " + numKnown);
		// logger.debug(getAgentID() + "> BO memory: " + this.getBOMemory());
		// }
		// LOGGING ->

		// SELECT AGENT BO
		if (dBuilder == AmoreDecision.DBUILDER_SELECTION.getDBuilder()) {
			int interactionsToInitiate = 0;

			switch (((Integer) paraMan.getParameter(Parameter.INTERACTION_CAPACITY_METHOD)).intValue()) {
				case 1: {
					interactionsToInitiate = (Integer) paraMan.getParameter(Parameter.NUM_INTERACTIONS_AT_ONCE);
					break;
				}
				case 2: {
					interactionsToInitiate = (Integer) paraMan.getParameter(Parameter.NUM_INTERACTIONS_AT_ONCE) < (Integer) paraMan
							.getParameter(Parameter.INTERACTION_CAPACITY)
							- numCurrentInteractions ? (Integer) paraMan
							.getParameter(Parameter.NUM_INTERACTIONS_AT_ONCE) : (Integer) paraMan
							.getParameter(Parameter.INTERACTION_CAPACITY)
							- numCurrentInteractions;
					break;
				}
				default: {
					throw new IllegalParameterException("Parameter for INTERACTION_CAPACITY_METHOD not supported");
				}
			}

			// add game for selected agents:
			if (getDecisionData(dBuilder).getDecider().getNumAvailableBOs() >= interactionsToInitiate) {
				for (LaraBehaviouralOption<ReSolEvoAgent> bo : getDecisionData(dBuilder).getDecider().getKSelectedBos(
						interactionsToInitiate)) {

					addPlayDecision(((SelectAgent) bo).getAlter(), true);
					addUsedBO((SelectAgent) bo);

					// <- LOGGING
					if (agentlogger != null) {
						agentlogger.log(AgentLevel.AGENT, this.getAgentID()
								+ ">AfterDecide: play decision added with: " + ((SelectAgent) bo).getAlter());
					}
					if (logger.isDebugEnabled())
						logger.debug(this.getAgentID() + "> order of k best " + bo.getKey());

					// LOGGING ->
				}
			} else {
				for (LaraBehaviouralOption<ReSolEvoAgent> bo : getDecisionData(dBuilder).getDecider().getKSelectedBos(
						getDecisionData(dBuilder).getDecider().getNumAvailableBOs())) {

					addPlayDecision(((SelectAgent) bo).getAlter(), true);
					addUsedBO((SelectAgent) bo);

					// <- LOGGING
					if (agentlogger != null) {
						agentlogger.log(AgentLevel.AGENT, this.getAgentID()
								+ ">AfterDecide: play decision added with: " + ((SelectAgent) bo).getAlter());
					}
					if (logger.isDebugEnabled())
						logger.debug(this.getAgentID() + "> order of k best " + bo.getKey());
					// LOGGING ->
				}
			}

			switch (((Integer) paraMan.getParameter(Parameter.AGENT_TRIGGERING)).intValue()) {
				case 3: {
					ArrayList<LaraDecisionBuilder> toRemove = new ArrayList<LaraDecisionBuilder>();
					for (LaraDecisionBuilder decisionBuilder : getPlayDecisionsActive()) {
						// process only initiated games here since all others are processed implicitly:
						if (((ReSoLEvoPlayDecisionBuilder) decisionBuilder).isInitiated()) {

							preprocess(decisionBuilder);
							decide(decisionBuilder);

							if (!(getDecisionData(decisionBuilder).getBo() instanceof AvoidSupportGame)) {
								// let the partner agent choose
								ReSolEvoAgent partner = getGamePartner((ReSoLEvoPlayDecisionBuilder) decisionBuilder);
								ReSoLEvoPlayDecisionBuilder partnerDBuilder = partner.getPassivePlayDecision(this);

								partner.preprocess(partnerDBuilder);
								partner.decide(partnerDBuilder);
								partner.processDecision(partnerDBuilder);

							}
							processDecision(decisionBuilder);

							if (!(getDecisionData(decisionBuilder).getBo() instanceof AvoidSupportGame)) {
								ReSolEvoAgent partner = getGamePartner((ReSoLEvoPlayDecisionBuilder) decisionBuilder);
								partner.clean(partner.getPassivePlayDecision(this));
							}
							toRemove.add(decisionBuilder);
						}
					}
					for (LaraDecisionBuilder decisionBuilder : toRemove) {
						clean(decisionBuilder);
					}
					break;
				}
				case 1:
				case 2: {
					// do nothing
					break;
				}
				default: {
					throw new IllegalParameterException("Parameter for AGENT_TRIGGERING not supported");
				}
			}
		}

		// PLAY GAME BO
		if (dBuilder instanceof ReSoLEvoPlayDecisionBuilder) {
			if (logger.isDebugEnabled())
				logger.debug(this.getAgentID() + "> afterDecide " + dBuilder);

			LaraDecisionData<ReSolEvoAgent, LaraBehaviouralOption<ReSolEvoAgent>> dData = getDecisionData(dBuilder);
			if (dData.getDecider().getNumAvailableBOs() > 0) {
				dData.setBo(dData.getDecider().getSelectedBo());

				// increase the number of interactions:
				// not initiated interactions are counted in interacts(), and AvoidSupportGame decisions may not be
				// counted
				if (((ReSoLEvoPlayDecisionBuilder) dBuilder).isInitiated()
						&& (!(dData.getBo() instanceof AvoidSupportGame))) {
					numCurrentInteractions++;
				}
			}
		}
	}

	/**
	 * This method is executed after every agent's afterDecide method was executed! > Update payoffs for play decisions.
	 * 
	 * @param dB Created by Sascha Holzhauer on 26.02.2010
	 */
	@Override
	public void processDecision(LaraDecisionBuilder dB) {
		logger.info(getAgentID() + "> processDecision");
		if (agentlogger != null) {
			agentlogger.log(AgentLevel.AGENT, getAgentID() + "> BO-Memory: " + this.getBOMemory().toString());
		}

		if (dB instanceof ReSoLEvoPlayDecisionBuilder) {
			ReSoLEvoPlayDecisionBuilder dBuilder = (ReSoLEvoPlayDecisionBuilder) dB;
			// the following only needs to be executed when the game was actually played and the agent was not rejected:
			// however, in case of rejection the decisionBuilder is actually removed.
			if (getDecisionData(dBuilder).getBo() instanceof PlaySupportGame) {
				AskKnownAgent selectBo;

				// add BO-memory entry if not in knowns (Pujol2005: payoff is updated which means agent becomes known):
				if (getLaraComp().getBOMemory().contains(getGamePartner(dBuilder).getAgentID())
						&& getLaraComp().getBOMemory().recall(getGamePartner(dBuilder).getAgentID()) instanceof AskKnownAgent) {
					ReSolEvoModel.increasenumUpdatedMems();
				}
				addKnownAgent(getGamePartner(dBuilder), 0.0, false);

				selectBo = (AskKnownAgent) getLaraComp().getBOMemory().recall(getGamePartner(dBuilder).getAgentID());

				// set actual payoffs:
				if (!(getDecisionData(dBuilder).getBo() instanceof AvoidSupportGame)) {
					updatePayoffs(dBuilder, selectBo);
				} else {
					// update memory when no game was played:
					Map<Class<? extends LaraGoal>, Double> utilities = selectBo.getModifiableUtilities();
					utilities.put(MaximizePayoffGoal.class, new Double(
							(utilities.get(MaximizePayoffGoal.class) + 0.0) / 2.0));

					getBOMemory().refresh(
							new AskKnownAgent(getGamePartner(dBuilder).getAgentID(), this, utilities,
									getGamePartner(dBuilder),
									// TODO make optional (important when changing memory capacity management) (?)
									((SelectAgent) getBOMemory().recall(getGamePartner(dBuilder).getAgentID()))
											.isFromME()));
				}
			}
		}
	}

	/**
	 * Updates the payoff values of corresponding Play SupportGame BOs. Pujol2005, 2.23
	 * 
	 * @param partner Created by Sascha Holzhauer on 05.03.2010
	 */
	private void updatePayoffs(ReSoLEvoPlayDecisionBuilder dBuilder, AskKnownAgent selectBo) {
		ReSolEvoModel.increaseNumGamesPlayed();
		Map<Class<? extends LaraGoal>, Double> utilities = selectBo.getModifiableUtilities();
		PlaySupportGame game = (PlaySupportGame) getDecisionData(dBuilder).getBo();

		// ReSolEvoPlayDecisionBuilder should always be associated with a PlaySupportGame (AvoidSupportGame is a
		// subclass):
		ReSolEvoAgent partner = game.getPartnerAgent();
		double benefit = getReceivingAdviceGain(partner);
		double cost = getGivingAdviseLoss(partner);

		// LOGGING ->
		logger.info(getAgentID() + "> Cooperation? " + getChosenPlayBo(partner).isCooperation() + " > " + partner);

		// <- LOGGING
		if (agentlogger != null) {
			agentlogger.log(AgentLevel.AGENT, getAgentID() + "> Game: " + this.getAgentID()
					+ (game.isCooperation() ? " cooperates, " : " defects, ") + partner.getAgentID()
					+ (partner.getChosenPlayBo(this).isCooperation() ? " cooperates" : " defects"));
		}

		if (game.isCooperation()) {
			// focal agent chose to cooperate:
			if (partner.getChosenPlayBo(this).isCooperation()) {
				// partner agent cooperates:
				utilities.put(MaximizePayoffGoal.class,
				// new entry: TODO check if this works (time stamp)!
						selectBo.getTimestamp() == LModel.getModel().getCurrentStep() ? (benefit - cost) :
						// partner known before:
								(utilities.get(MaximizePayoffGoal.class) + (benefit - cost)) / 2.0);

				// memory exchange:
				switch (((Integer) ReSolEvoParameterDB.getInstance().getParameter(Parameter.MEMORY_EXCHANGE))
						.intValue()) {
					case 1: {
						// UNDO
						gossip(partner);
						break;
					}
					case 2: {
						observe(partner);
						break;
					}
					case 3: {
						// TODO does order matter?
						observe(partner);
						gossip(partner);
						break;
					}
					default:
						throw new IllegalParameterException("Parameter for MEMORY_EXCHANGE not supported");
				}
			} else {
				// partner agent defects:
				utilities.put(MaximizePayoffGoal.class,
				// new entry:
						selectBo.getTimestamp() == LModel.getModel().getCurrentStep() ? -cost :
						// partner known before:
								((utilities.get(MaximizePayoffGoal.class) - cost) / 2.0));
			}
		} else {
			// focal agent defects:

			// <- LOGGING
			if (agentlogger != null) {
				agentlogger.log(AgentLevel.AGENT, this.getAgentID() + " > update Payoffs. Defects against "
						+ partner.getAgentID());
			}
			logger.debug(this.getAgentID() + " > update Payoffs. Defects against " + partner.getAgentID());
			// LOGGING ->

			if (partner.getChosenPlayBo(this).isCooperation()) {
				// partner agent cooperates (focal agent may not expect partner to cooperate next time):
				utilities.put(MaximizePayoffGoal.class, -benefit);
			} else {
				// partner agent defects:
				utilities.put(MaximizePayoffGoal.class, 0.0);
			}
		}

		logger.debug(">>> Outcome: " + utilities.get(MaximizePayoffGoal.class).floatValue());

		// the item may have been forgotten in meantime but needs to be added anyway since the payoff is updated:
		if (getLaraComp().getBOMemory().contains(selectBo.getKey())) {
			getLaraComp().getBOMemory().refresh(selectBo.getModifiedUtilitiesBO(utilities));
		} else {
			getLaraComp().getBOMemory().memorize(selectBo.getModifiedUtilitiesBO(utilities));
			numKnown++;
			if (selectBo.isFromME()) {
				numKnownME++;
			}
		}
	}

	@Override
	public void perceive(LaraDecisionBuilder dBuilder) {
		// nothing to do (not called)
	}

	/**
	 * Calculates the gain (G_(ij)) this agent receives when it gets advice from the given partner agent. (see
	 * Pujol2005, paragraph 2.11)
	 * 
	 * @param partnerAgent
	 * @return Created by Sascha Holzhauer on 24.02.2010
	 */
	public double getReceivingAdviceGain(ReSolEvoAgent partnerAgent) {
		return (1.0 - this.getAttractiveness()) * partnerAgent.getAttractiveness()
				* (Double) paraMan.getParameter(Parameter.ADVICE_BENEFIT);
	}

	/**
	 * Calculates the cost (L_(ij)) this agent experiences when it gives advice to the given partner agent. (see
	 * Pujol2005, paragraph 2.11)
	 * 
	 * @param partnerAgent
	 * @return Created by Sascha Holzhauer on 24.02.2010
	 */
	public double getGivingAdviseLoss(ReSolEvoAgent partnerAgent) {
		return (1.0 - partnerAgent.getAttractiveness()) * this.getAttractiveness()
				* (Double) paraMan.getParameter(Parameter.ADVICE_COST);
	}

	/**
	 * The method is invoked by partner agents to find out whether this agent may interact with them. Furthermore, here
	 * the interaction counter is increased and the play game decision builder is added.
	 * 
	 * See Pujol2005, section 2.20
	 * 
	 * @param partner the partner agent to interact with
	 * @return true if this agent is still able to interact Created by Sascha Holzhauer on 05.03.2010
	 */
	public boolean interacts(ReSolEvoAgent partner) {
		logger.debug(this.getAgentID() + "> check interact with " + partner.getAgentID());
		int numInterBefore = numCurrentInteractions;
		if (numCurrentInteractions < interaction_capacity) {
			if (knows(partner)) {
				SelectAgent selectBo = (SelectAgent) getLaraComp().getBOMemory().recall(partner.getAgentID());
				if (selectBo.getValue().get(MaximizePayoffGoal.class) > 0.0) { // TODO insert =
					numCurrentInteractions++;
					addPlayDecision(partner, false);
					// <- LOGGING
					if (agentlogger != null) {
						agentlogger.log(AgentLevel.AGENT, this.getAgentID() + ">Interact: play decision added with: "
								+ partner);
					}
					// LOGGING ->
					return true;
				} else {
					return false;
				}
			} else {
				switch (((Integer) ReSolEvoParameterDB.getInstance().getParameter(
						Parameter.INTERACTION_ACCEPTANCE_METHOD)).intValue()) {
					case 1: {
						// pseudo code version:
						numCurrentInteractions++;
						addPlayDecision(partner, false);
						// <- LOGGING
						if (agentlogger != null) {
							agentlogger.log(AgentLevel.AGENT, this.getAgentID()
									+ "> Interact: play decision added with: " + partner);
						}
						// LOGGING ->
						return true;
					}
					case 2: {
						// description version (only accept when agent is aware of):
						// does not work since mutual awareness is rare!!!
						if (isAwareOf(partner)) {
							numCurrentInteractions++;
							addPlayDecision(partner, false);
							return true;
						} else {
							return false;
						}
					}
					default: {
						throw new IllegalParameterException("Parameter for INTERACTION_ACCEPTANCE_METHOD not supported");
					}
				}
			}
		}
		// <- LOGGING
		if (agentlogger != null) {
			agentlogger.log(AgentLevel.AGENT, getAgentID() + "> Interactions before: " + numInterBefore + "; "
					+ (isAwareOf(partner) ? " is aware of " : (knows(partner) ? " knows " : " is not aware of "))
					+ partner.getAgentID() + " | Interactions now: " + numCurrentInteractions);
		}
		logger.info("Agent " + this + " refuses interaction with " + partner);

		// LOGGING ->
		return false;
	}

	/**
	 * Pujol2005, section 2.24 Created by Sascha Holzhauer on 09.03.2010
	 * 
	 * @param sender
	 */
	public void gossip(ReSolEvoAgent sender) {

		ReSolEvoModel.increasenumExchangedMmeories();

		ReSolEvoAgent bestPartner = null;
		SelectAgent selectBo = null;
		ArrayList<ReSolEvoAgent> bestSelectAgents = new ArrayList<ReSolEvoAgent>();
		double bestExPayoff = Double.NEGATIVE_INFINITY;

		// find 'best' partner from sender's memory that it _knows_:
		for (LaraBehaviouralOption bo : sender.getLaraComp().getBOMemory().recallAllMostRecent()) {
			if (bo instanceof AskKnownAgent) {
				selectBo = (AskKnownAgent) bo;
				if (selectBo.getAlter() != this)
					if (selectBo.getValue().get(MaximizePayoffGoal.class) > bestExPayoff) {
						bestSelectAgents.clear();
						bestExPayoff = selectBo.getValue().get(MaximizePayoffGoal.class);
						bestPartner = selectBo.getAlter();
					} else if (selectBo.getValue().get(MaximizePayoffGoal.class) == bestExPayoff) {
						bestSelectAgents.add(selectBo.getAlter());
					}
			}
		}
		if (bestSelectAgents.size() > 0) {
			bestPartner = bestSelectAgents.get(RandomHelper.nextIntFromTo(0, bestSelectAgents.size() - 1));
		}

		// Pujol2005, section 2.25:
		if (bestPartner != null) {
			// bestPartner exists:
			if (getLaraComp().getBOMemory().contains(bestPartner.getAgentID() + ExploreAgent.EXPLORE_PREFIX)) {
				// bestPartner is noticed:

				SelectAgent exchangedPartner = (SelectAgent) getLaraComp().getBOMemory().recall(
						bestPartner.getAgentID() + ExploreAgent.EXPLORE_PREFIX);
				Map<Class<? extends LaraGoal>, Double> utilities = exchangedPartner.getModifiableUtilities();

				if (((SelectAgent) getLaraComp().getBOMemory().recall(
						bestPartner.getAgentID() + ExploreAgent.EXPLORE_PREFIX)).isFromME()) {
					// best partner is noticed from memory exchange:
					// only required when Parameter.MEMORY_EXCHANGE_CONSEQUENCE is set to 2 or 3

					utilities.put(MaximizePayoffGoal.class, new Double(
							(selectBo.getValue().get(MaximizePayoffGoal.class) + exchangedPartner.getValue().get(
									MaximizePayoffGoal.class)) / 2.0));
					getLaraComp().getBOMemory().refresh(exchangedPartner.getModifiedBO(this, utilities));
				} else {
					// best partner is noticed , but not from memory exchange:
					utilities.put(MaximizePayoffGoal.class, new Double(selectBo.getValue()
							.get(MaximizePayoffGoal.class)));
					getLaraComp().getBOMemory().refresh(exchangedPartner.getModifiedBO(this, utilities));
				}
			} else if (getLaraComp().getBOMemory().contains(bestPartner.getAgentID())
					&& ((SelectAgent) getLaraComp().getBOMemory().recall(bestPartner.getAgentID())).isFromME()) {
				// best parter is known from memory exchange:
				SelectAgent exchangedPartner = (SelectAgent) getLaraComp().getBOMemory().recall(
						bestPartner.getAgentID());
				Map<Class<? extends LaraGoal>, Double> utilities = exchangedPartner.getModifiableUtilities();
				utilities.put(MaximizePayoffGoal.class, new Double(
						(selectBo.getValue().get(MaximizePayoffGoal.class) + exchangedPartner.getValue().get(
								MaximizePayoffGoal.class)) / 2.0));
				getLaraComp().getBOMemory().refresh(exchangedPartner.getModifiedBO(this, utilities));
			} else {
				// agent is not noticed and not known:
				switch (((Integer) ReSolEvoParameterDB.getInstance()
						.getParameter(Parameter.MEMORY_EXCHANGE_CONSEQUENCE)).intValue()) {
					case 1: {
						if (logger.isDebugEnabled()) {
							logger.debug("Memory Exchange (Gossip): " + sender + " ---" + bestPartner + "--> " + this);
						}
						addKnownAgent(bestPartner, bestExPayoff, true);
						break;
					}
					case 2: {
						addNoticedAgent(bestPartner, true);
						break;
					}
					case 3: {
						addKnownAgent(bestPartner, bestExPayoff, true);
						break;
					}
					default:
						throw new IllegalParameterException(
								"Parameter for MEMORY_EXCHANGE_CONSEQUENCE is not supported");
				}
			}

		}
	}

	/**
	 * Pujol2005, section 2.24
	 * 
	 * @param observed agent
	 * 
	 *            Created by Sascha Holzhauer on 09.03.2010
	 */
	public void observe(ReSolEvoAgent observed) {

		if (logger.isDebugEnabled()) {
			logger.debug(this + " observes " + observed);
		}

		ReSolEvoModel.increasenumExchangedMmeories();

		ReSolEvoAgent exchanged = null;
		ArrayList<ReSolEvoAgent> potPartners = new ArrayList<ReSolEvoAgent>(observed.getDecisionDataSize());
		// scan decisionData for appropriate Bos:
		for (LaraDecisionData<ReSolEvoAgent, LaraBehaviouralOption<ReSolEvoAgent>> dd : observed
				.getDecisionDataIterable()) {
			if (dd.getdBuilder() instanceof ReSoLEvoPlayDecisionBuilder
					&& (!potPartners.contains(((PlaySupportGame) dd.getBo()).getPartnerAgent()))) {
				potPartners.add(((PlaySupportGame) dd.getBo()).getPartnerAgent());
			}
		}
		do {
			int rand = RandomHelper.getUniform().nextIntFromTo(0, potPartners.size() - 1);
			exchanged = potPartners.get(rand);
		} while (exchanged == this && potPartners.size() > 1); // must at least contain 2 to allow another loop!
		if (exchanged != this) {

			// from pseudo code: exchange agents are added to knowns since time > 0 because of current interaction of
			// observed:
			switch (((Integer) ReSolEvoParameterDB.getInstance().getParameter(Parameter.MEMORY_EXCHANGE_CONSEQUENCE))
					.intValue()) {
				// TODO check whether choice for bestExPayoff is appropriate!
				case 1: {
					// TODO check what to do when entry was deleted from memory in meantime!
					addKnownAgent(exchanged, observed.getBOMemory().contains(exchanged.getAgentID()) ? (observed
							.getBOMemory().recall(exchanged.getAgentID()).getValue().get(MaximizePayoffGoal.class))
							.doubleValue() : 0.0, true);
					break;
				}
				case 2: {
					addNoticedAgent(exchanged, true);
					break;
				}
				case 3: {
					addNoticedAgent(exchanged, true);
					break;
				}
				default:
					throw new IllegalParameterException("Parameter for MEMORY_EXCHANGE_CONSEQUENCE is not supported");
			}
		}
	}

	/**
	 * Checks whether the given partner agent is in the memory of known. (according to pseudo code - alternative:
	 * network neighbourhood which depends on network's directedness)
	 * 
	 * @param partner
	 * @return true if the given agent is a neighbour in the known network Created by Sascha Holzhauer on 05.03.2010
	 */
	private boolean knows(ReSolEvoAgent partner) {
		return this.getLaraComp().getBOMemory().contains(partner.getAgentID());
	}

	/**
	 * If the given agent is noticed be this agent but not known!
	 * 
	 * @param partner
	 * @return Created by Sascha Holzhauer on 08.03.2010
	 */
	private boolean isAwareOf(ReSolEvoAgent partner) {
		return this.getLaraComp().getBOMemory().contains(partner.getAgentID() + ExploreAgent.EXPLORE_PREFIX);
	}

	/**
	 * @return Created by Sascha Holzhauer on 26.02.2010
	 */
	public int getAllNeigbours() {
		return numKnown + numNoticed;
	}

	/**
	 * Returns the number of agents to explore
	 * 
	 * @return the number of agents this agent is connected to in the noticed network Created by Sascha Holzhauer on
	 *         03.05.2010
	 */
	public int getNumNoticed() {
		return numNoticed;
	}

	/**
	 * @return the number of agents this agent is connected to in the knowns network without those that are solely known
	 *         through memory exchange Created by Sascha Holzhauer on 18.08.2010
	 */
	public int getNumKnownWithoutME() {
		return numKnown - numKnownME;
	}

	/**
	 * Used for analysing the resulting network!
	 * 
	 * @return this agent's degree in the known's network Created by Sascha Holzhauer on 20.04.2010
	 */
	public int getKnownNetworkDegree() {

		// TODO
		// ReSolEvoNetwork net = (ReSolEvoNetwork)
		// this.getLaraNetworkComp().getNetwork(ReSolEvoModel.Network.NETWORK_KNOWN.toString()).;
		// net.initNetwork();
		// if (net.getDegree(this) != numKnown) {
		// logger.warn("Num of Known (" + numKnown + ") entspricht nicht network degree (" + net.getDegree(this) + ")");
		// }
		logger.debug("Numer of known: " + numKnown);
		return ((ReSolEvoNetwork) ((DefaultLRsNetwork) this.getLaraNetworkComp().getNetwork(
				ReSolEvoModel.Network.NETWORK_KNOWN.toString())).getNetwork()).getUndirected().degree(this);
	}

	/**
	 * @param partner
	 * @return the chosenPlayBo with the given partner
	 */
	private PlaySupportGame getChosenPlayBo(ReSolEvoAgent partner) {
		// <- LOGGING
		if (agentlogger != null) {
			agentlogger.log(AgentLevel.AGENT, getAgentID() + "> getChosenPlayBO for " + partner + "\n"
					+ getDecisionDataIterable());
		}
		// LOGGING ->

		for (LaraDecisionData<ReSolEvoAgent, LaraBehaviouralOption<ReSolEvoAgent>> dData : getDecisionDataIterable()) {
			if (dData.getBo() instanceof PlaySupportGame) {
				if (((PlaySupportGame) dData.getBo()).getPartnerAgent() == partner) {
					return (PlaySupportGame) dData.getBo();
				}
			}
		}
		throw new IllegalStateException(this + "> BO for partner agent " + partner + " not found!");
	}

	/**
	 * 
	 * "When adding a key-value pair, the value may already exist in the map against a different key. That mapping is
	 * removed, to ensure that the value only occurs once in the inverse map." !!!
	 * 
	 * @param partner Created by Sascha Holzhauer on 08.03.2010
	 * @param initiated has this agent initiated the interaction?
	 */
	public void addPlayDecision(ReSolEvoAgent partner, boolean initiated) {

		if (initiated) {
			gameDecisionsActive.put(new ReSoLEvoPlayDecisionBuilder(partner.getAgentID(),
					new LDecisionHeuristicComponent_MaxLineTotalRandomAtTie(), initiated), partner);
			logger.info("Agent " + this + " found partner " + partner);
		} else {
			ReSoLEvoPlayDecisionBuilder dBuilder = new ReSoLEvoPlayDecisionBuilder(partner.getAgentID(),
					new LDecisionHeuristicComponent_MaxLineTotalRandomAtTie(), initiated);
			gameDecisionsPassive.put(dBuilder, partner);
			switch (((Integer) paraMan.getParameter(Parameter.AGENT_TRIGGERING)).intValue()) {
				case 2: {
					preprocess(dBuilder);
					break;
				}
				default:
					break;
			}
		}
		if (logger.isDebugEnabled()) {
			logger.debug(this + "> Play decision added for " + partner + "\n" + gameDecisionsPassive);
		}
	}

	/**
	 * @return Created by Sascha Holzhauer on 08.03.2010
	 */
	public Set<ReSoLEvoPlayDecisionBuilder> getPlayDecisionsActive() {
		return gameDecisionsActive.keySet();
	}

	/**
	 * @return Created by Sascha Holzhauer on 08.03.2010
	 */
	public Set<ReSoLEvoPlayDecisionBuilder> getPlayDecisionsPassive() {
		return gameDecisionsPassive.keySet();
	}

	/**
	 * Removes the specified dBuilder from the map of current play decisions.
	 * 
	 * @param dBuilder Created by Sascha Holzhauer on 05.05.2010
	 */
	public void removeActiveGameDecision(ReSoLEvoPlayDecisionBuilder dBuilder) {
		gameDecisionsActive.remove(dBuilder);
		if (logger.isDebugEnabled()) {
			logger.debug(getAgentID() + ">>> Play decision removed for " + dBuilder);
		}
	}

	/**
	 * @param partner
	 * @return Created by Sascha Holzhauer on 08.03.2010
	 */
	public ReSoLEvoPlayDecisionBuilder getActivePlayDecision(ReSolEvoAgent partner) {
		return gameDecisionsActive.getKey(partner);
	}

	/**
	 * @param partner
	 * @return Created by Sascha Holzhauer on 08.03.2010
	 */
	public ReSoLEvoPlayDecisionBuilder getPassivePlayDecision(ReSolEvoAgent partner) {
		return gameDecisionsPassive.getKey(partner);
	}

	/**
	 * Return the partner to the given game.
	 * 
	 * @param dBuilder
	 * @return Created by Sascha Holzhauer on 29.04.2010
	 */
	public ReSolEvoAgent getGamePartner(ReSoLEvoPlayDecisionBuilder dBuilder) {
		if (dBuilder.isInitiated())
			return gameDecisionsActive.get(dBuilder);
		else
			return gameDecisionsPassive.get(dBuilder);
	}

	// registered only for FORGOTTEN-Events and REFRESHED_ITEM_FORGOTTEN...
	@Override
	public void memoryPropertyEventOccured(MemoryPropertyEvent event, LaraProperty<?> property) {

		if (event == MemoryPropertyEvent.REFRESHED_ITEM_FORGOTTEN) {
			// correct auto-removed memory properties:
			if (property instanceof AskKnownAgent) {
				logger.debug("Known BO (" + property + ") refreshed in memory (" + this + ")");
				AskKnownAgent prop = (AskKnownAgent) property;
				numKnown++;
				if (prop.isFromME()) {
					numKnownME++;
				}
			}
			if (property instanceof ExploreAgent) {
				logger.debug("Noticed BO (" + property + ") refreshed in memory (" + this + ")");
				numNoticed++;
			}
		}

		if (event == MemoryPropertyEvent.ITEM_FORGOTTEN) {
			if (property instanceof AskKnownAgent) {
				if (logger.isDebugEnabled()) {
					logger.debug("Known BO (" + property + ") removed from memory (" + this + ")");
				}
				AskKnownAgent prop = (AskKnownAgent) property;

				numKnown--;
				if (numKnown < 0) {
					logger.warn("Num of Known is negative! (" + numKnown + "/event: " + event + ")");
				}

				if (prop.isFromME()) {
					numKnownME--;
					if (numKnownME < 0) {
						logger.warn("Num of KnownME is negative! (" + numKnownME + ")");
					}
				}
			}
			if (property instanceof ExploreAgent) {
				numNoticed--;
				if (numNoticed < 0) {
					logger.warn("Num of Noticed is negative! (" + numNoticed + ")");
				}
			}
		}
	}

	/**
	 * A comparator that compares {@link ReSolEvoAgent}s according to its agent id.
	 * 
	 * @return Created by Sascha Holzhauer on 08.04.2010
	 */
	public static Comparator<ReSolEvoAgent> getComparator() {
		return new Comparator<ReSolEvoAgent>() {
			@Override
			public int compare(ReSolEvoAgent o1, ReSolEvoAgent o2) {
				return o1.getAgentID().compareTo(o2.getAgentID());
			}
		};
	}

	/**
	 * @return the current parameter id Created by Sascha Holzhauer on 21.04.2010
	 */
	public int getParamID() {
		return ((Integer) ReSolEvoParameterDB.getInstance().getParameter(Parameter.PARAMS_ID)).intValue();
	}

	/**
	 * This should be save since the agentID is unique!
	 * 
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		int result = 17;
		result = 31 * result + (getAgentID() == null ? 0 : getAgentID().hashCode());
		return result;
	}

	@Override
	public int compareTo(ReSolEvoAgent o) {
		return getAgentID().compareTo(o.getAgentID());
	}

	/**
	 * Adds a noticed agent to attractiveness memory and BO-memory and connects it in noticed network. Does nothing when
	 * ExploreAgent already exists. System
	 * 
	 * @param alter
	 * @param fromME Created by Sascha Holzhauer on 05.05.2010
	 */
	public void addNoticedAgent(ReSolEvoAgent alter, boolean fromME) {
		if (!getLaraComp().getBOMemory().contains(alter.getAgentID() + ExploreAgent.EXPLORE_PREFIX)) {
			this.getLaraComp().getBOMemory().memorize(
					new ExploreAgent(alter.getAgentID() + ExploreAgent.EXPLORE_PREFIX, this, alter, fromME));
			numNoticed++;
		}
	}

	/**
	 * Adds a known agent to attractiveness memory and BO-memory and deletes ExploreAgent-BO if exists. Does nothing in
	 * the according memory if the entry already exists in there.
	 * 
	 * TODO in case of Explore BO exists the counter needs to be increased, at least...
	 * 
	 * @param alter
	 * @param exPayoff
	 * @param fromME Created by Sascha Holzhauer on 05.05.2010
	 */
	public void addKnownAgent(ReSolEvoAgent alter, double exPayoff, boolean fromME) {
		if (!getLaraComp().getBOMemory().contains(alter.getAgentID())) {

			Map<Class<? extends LaraGoal>, Double> utilities = new HashMap<Class<? extends LaraGoal>, Double>();
			utilities.put(MaximizePayoffGoal.class, new Double(exPayoff));
			getLaraComp().getBOMemory().memorize(new AskKnownAgent(alter.getAgentID(), this, utilities, alter, fromME));
			numKnown++;
			if (fromME) {
				numKnownME++;
			}
			if (getLaraComp().getBOMemory().contains(alter.getAgentID() + ExploreAgent.EXPLORE_PREFIX)) {
				getLaraComp().getBOMemory().forgetAll(alter.getAgentID() + ExploreAgent.EXPLORE_PREFIX);
			}
		}
	}

	/**
	 * Created by Sascha Holzhauer on 08.03.2010
	 */
	@Override
	public void customClean(LaraDecisionBuilder dBuilder) {

		// <- LOGGING
		if (logger.isDebugEnabled()) {
			logger.debug("Clean: Number of Kown: " + numKnown);
			logger.debug(getAgentID() + "> BO memory: " + this.getBOMemory());
		}
		// LOGGING ->

		if (dBuilder == ReSolEvoModel.AmoreDecision.DBUILDER_SELECTION.getDBuilder()) {
			usedBOs.clear();
		}

		else {
			// delete PlaySupport LaraDecisionData objects:
			ArrayList<LaraDecisionBuilder> toRemove = new ArrayList<LaraDecisionBuilder>(getDecisionDataSize());
			for (LaraDecisionData<ReSolEvoAgent, LaraBehaviouralOption<ReSolEvoAgent>> dData : getDecisionDataIterable()) {
				if (logger.isDebugEnabled()) {
					logger.debug(this + "> Check DecisionData of " + dData.getdBuilder().getId() + " to remove.");
				}
				if (dData.getdBuilder() instanceof ReSoLEvoPlayDecisionBuilder) {
					if (logger.isDebugEnabled()) {
						logger.debug(this + "> Added DecisionData of " + dData.getdBuilder().getId() + " to remove.");
					}
					toRemove.add(dData.getdBuilder());
				}
			}
			for (LaraDecisionBuilder dB : toRemove) {
				removeDecisionData(dB);
			}

			gameDecisionsActive.clear();
			gameDecisionsPassive.clear();

			// reset counter:
			numCurrentInteractions = 0;
		}
	}

	@Override
	public Object getNetworkMeasureObject(ContextJungNetwork<? extends NetworkMeasureSupport> network,
			MeasureDescription key) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public void setNetworkMeasureObject(ContextJungNetwork<? extends NetworkMeasureSupport> network,
			MeasureDescription key, Object value) {
		// TODO Auto-generated method stub
		
	}	
}